<?php

declare(strict_types=1);

namespace Drupal\sm\Transport\DrupalSql;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Database\Connection as CoreConnection;
use Drupal\Core\Database\DatabaseException;
use Drupal\mysql\Driver\Database\mysql\Connection as MySqlConnection;
use Symfony\Component\Messenger\Exception\InvalidArgumentException;
use Symfony\Component\Messenger\Exception\TransportException;

/**
 * Connection to the Drupal database for the SQL transport.
 *
 * @phpstan-type DrupalSqlConfiguration array{table_name: string, queue_name: string, redeliver_timeout: positive-int, auto_setup: bool}
 * @phpstan-type DrupalSqlEnvelope array{id: MessageId, body: string, headers: DrupalSqlEnvelopeHeaders, queue_name: string, created_at: string, available_at: string, delivered_at: string|null}
 * @phpstan-type DrupalSqlEnvelopeEncoded array{id: string, body: string, headers: string, queue_name: string, created_at: string, available_at: string, delivered_at: string|null}
 * @phpstan-type DrupalSqlEnvelopeHeaders string[]
 *
 * @internal
 */
final class Connection {

  /**
   * Date format for fields created_at, available_at, delivered_at.
   */
  private const DATE_FORMAT = 'Y-m-d H:i:s';

  /**
   * A set of default options for the connection.
   *
   * @const DrupalSqlConfiguration
   */
  protected const DEFAULT_OPTIONS = [
    'table_name' => 'sm_messages',
    // Queue name is a column, and may be processed independently by consume
    // command.
    'queue_name' => 'default',
    'redeliver_timeout' => 3600,
    'auto_setup' => TRUE,
  ];

  private bool $autoSetup;

  /**
   * Whether to perform MySQL cleanup.
   *
   * On MySQL, we mark a message as delivered by setting the `delivered_at`
   * field to a far future date (9999-12-31 23:59:59) instead of deleting it to
   * avoid deadlocks when multiple workers are processing messages.
   * To avoid a lot of DELETE queries during the `get()` operation, we set this
   * flag to TRUE when we acknowledge or reject a message, and then we perform
   * the cleanup in the `get()` method only if the flag is set.
   *
   * @var bool
   *
   * @see https://github.com/symfony/symfony/pull/45888
   * @see https://github.com/symfony/symfony/pull/54606
   */
  private bool $doMysqlCleanup = FALSE;

  /**
   * @phpstan-param DrupalSqlConfiguration $configuration
   */
  public function __construct(
    private readonly CoreConnection $connection,
    protected array $configuration,
    private readonly TimeInterface $time,
    private readonly MessageTableTracker $messageTableTracker,
  ) {
    // @phpstan-ignore-next-line
    $this->configuration = \array_replace_recursive(static::DEFAULT_OPTIONS, $configuration);
    $this->autoSetup = (bool) $this->configuration['auto_setup'];
  }

  /**
   * @phpstan-param array<string, mixed> $options
   * @phpstan-return DrupalSqlConfiguration
   */
  public static function buildConfiguration(#[\SensitiveParameter] string $dsn, array $options = []): array {
    if (FALSE === \str_starts_with($dsn, 'drupal-sql://')) {
      throw new InvalidArgumentException('DSN must start with "drupal-sql://"');
    }

    if (FALSE === \str_starts_with($dsn, 'drupal-sql://default')) {
      throw new InvalidArgumentException('Only the `default` connection is supported at this time.');
    }

    $params = \parse_url($dsn);
    if (FALSE === $params) {
      throw new InvalidArgumentException('The given Drupal SQL DSN is invalid.');
    }

    $query = [];
    if (isset($params['query'])) {
      \parse_str($params['query'], $query);
    }

    $configuration = [];
    $configuration += $query + $options + static::DEFAULT_OPTIONS;
    $configuration['auto_setup'] = \filter_var($configuration['auto_setup'], \FILTER_VALIDATE_BOOL);

    // Check for extra keys in options.
    $optionsExtraKeys = \array_diff(\array_keys($options), \array_keys(static::DEFAULT_OPTIONS));
    if (0 < \count($optionsExtraKeys)) {
      throw new InvalidArgumentException(\sprintf('Unknown option found: [%s]. Allowed options are [%s].', \implode(', ', $optionsExtraKeys), \implode(', ', \array_keys(static::DEFAULT_OPTIONS))));
    }

    // Check for extra keys in DSN query.
    $queryExtraKeys = \array_diff(\array_keys($query), \array_keys(static::DEFAULT_OPTIONS));
    if (0 < \count($queryExtraKeys)) {
      throw new InvalidArgumentException(\sprintf('Unknown option found in DSN: [%s]. Allowed options are [%s].', \implode(', ', $queryExtraKeys), \implode(', ', \array_keys(static::DEFAULT_OPTIONS))));
    }

    /** @var DrupalSqlConfiguration */
    return $configuration;
  }

  /**
   * Sends the given Envelope to the Drupal SQL database.
   *
   * If a delay is specified, the message will be available after the delay
   * period.
   *
   * When autoSetup is enabled, an attempt to create the table will be made
   * on the first failure or row insert.
   *
   * @param string $body
   *   The serialized body of the message.
   * @param DrupalSqlEnvelopeHeaders $headers
   *   The headers of the message.
   * @param int $delay
   *   The delay in milliseconds before the message is available.
   *
   * @throws \Exception
   */
  public function send(string $body, array $headers, int $delay = 0): MessageId {
    $now = new \DateTimeImmutable('@' . $this->time->getCurrentTime());
    $availableAt = $now->modify(\sprintf('%+d seconds', $delay / 1e3));
    $fields = [
      'body' => $body,
      'headers' => \json_encode($headers),
      'queue_name' => $this->configuration['queue_name'],
      'created_at' => $now->format(static::DATE_FORMAT),
      'available_at' => $availableAt->format(static::DATE_FORMAT),
    ];

    $insert = function () use ($fields): MessageId {
      return new MessageId(
        (string) $this->connection
          ->insert($this->configuration['table_name'])
          ->fields($fields)
          ->execute(),
      );
    };

    $transaction = $this->connection->startTransaction();
    try {
      return $insert();
    }
    catch (\Exception $e) {
      $transaction->rollBack();

      // If there was an exception, try to create the table.
      if ($this->ensureTableExists()) {
        $transaction = $this->connection->startTransaction();
        try {
          return $insert();
        }
        catch (\Exception $e) {
          $transaction->rollBack();
        }
      }

      // If the exception happened for another reason than the missing table,
      // propagate the exception.
      throw $e;
    }
  }

  /**
   * Retrieves the next available message from the configured table.
   *
   * This method fetches the next message that is available for processing
   * from the database. It marks the message as delivered by updating the
   * `delivered_at` timestamp.
   *
   * If an exception occurs during the retrieval process, it attempts to
   * ensure the table exists and retries the operation.
   *
   * @return DrupalSqlEnvelope|null
   *   The decoded message as an associative array, or NULL if no message is
   *   available.
   *
   * @throws \Exception
   */
  public function get(): ?array {
    $get = function (): ?array {
      if ($this->doMysqlCleanup) {
        $this
          ->connection
          ->delete($this->configuration['table_name'])
          ->condition('delivered_at', '9999-12-31 23:59:59')
          ->execute();
        $this->doMysqlCleanup = FALSE;
      }

      $now = new \DateTimeImmutable('@' . $this->time->getCurrentTime());
      $redeliverLimit = $now->modify(\sprintf('-%d seconds', $this->configuration['redeliver_timeout']));

      $query = $this->connection->select($this->configuration['table_name']);

      $deliveredAtCondition = $query->orConditionGroup()
        ->isNull('delivered_at')
        ->condition('delivered_at', $redeliverLimit->format(static::DATE_FORMAT), '<=');

      $query
        ->fields($this->configuration['table_name'])
        ->orderBy('available_at')
        ->condition('queue_name', $this->configuration['queue_name'])
        ->condition('available_at', $now->format(static::DATE_FORMAT), '<=')
        ->condition($deliveredAtCondition)
        ->range(0, 1);

      // Drupal's typehint for fetchAssoc is incorrect: Only false, not bool.
      /** @var DrupalSqlEnvelopeEncoded|false $message */
      $message = $query->execute()?->fetchAssoc() ?? throw new \LogicException('query is not valid.');

      if ($message === FALSE) {
        return NULL;
      }

      $drupalSqlEnvelope = static::decodeEnvelope($message);

      $this
        ->connection
        ->update($this->configuration['table_name'])
        ->fields(['delivered_at' => $now->format(static::DATE_FORMAT)])
        ->condition('id', $drupalSqlEnvelope['id']->id)
        ->execute();

      return $drupalSqlEnvelope;
    };

    $transaction = $this->connection->startTransaction();
    try {
      return $get();
    }
    catch (\Exception $e) {
      $transaction->rollBack();

      // If there was an exception, try to create the table.
      if ($this->ensureTableExists()) {
        $transaction = $this->connection->startTransaction();
        try {
          return $get();
        }
        catch (\Exception $e) {
          $transaction->rollBack();
        }
      }

      // If the exception happened for other reason than the missing table,
      // propagate the exception.
      throw $e;
    }
  }

  /**
   * Acknowledges that the message was handled.
   *
   * @return bool
   *   TRUE if the message was successfully acknowledged, or FALSE if not.
   *
   * @throws \Symfony\Component\Messenger\Exception\TransportException
   */
  public function ack(MessageId $id): bool {
    try {
      // If the connection is MySQL, we update the `delivered_at` field to a far
      // future date instead of deleting the message to avoid deadlocks when
      // multiple workers are processing messages.
      if ($this->connection instanceof MySqlConnection) {
        $updated = $this
          ->connection
          ->update($this->configuration['table_name'])
          ->fields(['delivered_at' => '9999-12-31 23:59:59'])
          ->condition('id', $id->id)
          ->execute() > 0;

        if ($updated) {
          $this->doMysqlCleanup = TRUE;
        }

        return $updated;
      }

      return $this
        ->connection
        ->delete($this->configuration['table_name'], ['id' => $id->id])
        ->execute() > 0;
    }
    catch (\Exception $exception) {
      throw new TransportException($exception->getMessage(), previous: $exception);
    }
  }

  /**
   * Rejects the message with the given ID.
   *
   * This method updates the `delivered_at` timestamp to indicate that the
   * message has been rejected and will not be processed.
   *
   * @return bool
   *   TRUE if the message was successfully rejected, or FALSE if it was not.
   *
   * @throws \Symfony\Component\Messenger\Exception\TransportException
   */
  public function reject(MessageId $id): bool {
    try {
      // If the connection is MySQL, we update the `delivered_at` field to a far
      // future date instead of deleting the message to avoid deadlocks when
      // multiple workers are processing messages.
      if ($this->connection instanceof MySqlConnection) {
        $updated = $this
          ->connection
          ->update($this->configuration['table_name'])
          ->fields(['delivered_at' => '9999-12-31 23:59:59'])
          ->condition('id', $id->id)
          ->execute() > 0;

        if ($updated) {
          $this->doMysqlCleanup = TRUE;
        }

        return $updated;
      }

      return $this
        ->connection
        ->delete($this->configuration['table_name'], ['id' => $id->id])
        ->execute() > 0;
    }
    catch (\Exception $exception) {
      throw new TransportException($exception->getMessage(), previous: $exception);
    }
  }

  /**
   * Ensures that the messages table exists in the database.
   *
   * An attempt to create the table may be made if it does not exist.
   *
   * @return bool
   *   Whether the table already exists or was successfully created. FALSE is
   *   returned if table creation fails.
   */
  protected function ensureTableExists(): bool {
    if (FALSE === $this->autoSetup) {
      return $this->connection->schema()->tableExists($this->configuration['table_name']);
    }

    try {
      $this->connection
        ->schema()
        ->createTable($this->configuration['table_name'], Schema::schemaDefinition());

      $this->messageTableTracker->addMessageTable($this->configuration['table_name']);
    }
    // If another process has already created the messenger table, attempting
    // to recreate it will throw an exception. In this case just catch the
    // exception and do nothing. (This is the same technique used by other
    // Drupal systems: cache, batch, flood, lock, etc).
    catch (DatabaseException) {
    }
    catch (\Exception) {
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Decodes the headers of the given message envelope.
   *
   * This method takes an associative array representing a message envelope
   * and decodes the JSON-encoded headers into an associative array.
   *
   * @param DrupalSqlEnvelopeEncoded $encodedEnvelope
   *   The message envelope with JSON-encoded headers.
   *
   * @return DrupalSqlEnvelope
   *   The message envelope with decoded headers.
   */
  private static function decodeEnvelope(array $encodedEnvelope): array {
    /** @var DrupalSqlEnvelopeHeaders $headers */
    $headers = \json_decode($encodedEnvelope['headers'], TRUE);
    $encodedEnvelope['headers'] = $headers;
    $encodedEnvelope['id'] = new MessageId($encodedEnvelope['id']);

    return $encodedEnvelope;
  }

}
