<?php

declare(strict_types=1);

namespace Drupal\sm\Transport\DrupalSql;

use Drupal\sm\Transport\DrupalSql\Stamp\ReceivedStamp;
use Symfony\Component\Messenger\Envelope;
use Symfony\Component\Messenger\Exception\LogicException;
use Symfony\Component\Messenger\Exception\MessageDecodingFailedException;
use Symfony\Component\Messenger\Exception\TransportException;
use Symfony\Component\Messenger\Stamp\TransportMessageIdStamp;
use Symfony\Component\Messenger\Transport\Receiver\ReceiverInterface;
use Symfony\Component\Messenger\Transport\Serialization\SerializerInterface;

/**
 * A receiver for the Drupal SQL transport.
 *
 * @phpstan-import-type DrupalSqlEnvelope from Connection
 * @internal
 */
final class Receiver implements ReceiverInterface {

  /**
   * @param \Drupal\sm\Transport\DrupalSql\Connection $connection
   *   The transport connection.
   * @param \Symfony\Component\Messenger\Transport\Serialization\SerializerInterface $serializer
   *   A serializer to decode the serialized envelope from Sender::send().
   */
  public function __construct(
    private readonly Connection $connection,
    private readonly SerializerInterface $serializer,
  ) {
  }

  /**
   * {@inheritdoc}
   */
  public function get(): iterable {
    try {
      $envelope = $this->connection->get();
    }
    catch (\Exception $exception) {
      throw new TransportException($exception->getMessage(), previous: $exception);
    }

    if (NULL === $envelope) {
      return [];
    }

    return [$this->createEnvelopeFromData($envelope)];
  }

  public function ack(Envelope $envelope): void {
    $this->connection->ack(static::findReceivedStamp($envelope)->id);
  }

  public function reject(Envelope $envelope): void {
    $this->connection->reject(static::findReceivedStamp($envelope)->id);
  }

  /**
   * Creates an Envelope object from the provided data array.
   *
   * This method decodes the message data using the serializer and constructs
   * an Envelope object. If the message decoding fails, the message is
   * rejected.
   * The Envelope is then returned with the appropriate stamps.
   *
   * @phpstan-param DrupalSqlEnvelope $data
   *
   * @throws \Symfony\Component\Messenger\Exception\MessageDecodingFailedException
   *   Thrown if message decoding fails.
   */
  private function createEnvelopeFromData(array $data): Envelope {
    try {
      $envelope = $this->serializer->decode([
        'body' => $data['body'],
        'headers' => $data['headers'],
      ]);
    }
    catch (MessageDecodingFailedException $exception) {
      $this->connection->reject($data['id']);

      throw $exception;
    }

    return $envelope->with(
      ReceivedStamp::createFromData($data),
      new TransportMessageIdStamp($data['id']),
    );
  }

  /**
   * @throws \Symfony\Component\Messenger\Exception\LogicException
   *   Thrown if no stamp is not found. This should never happen since all
   *   messages that enter Drupal SQL transports get this stamp.
   */
  private static function findReceivedStamp(Envelope $envelope): ReceivedStamp {
    return $envelope->last(ReceivedStamp::class) ?? throw new LogicException(\sprintf('No %s found on the Envelope.', ReceivedStamp::class));
  }

}
