<?php

declare(strict_types=1);

namespace Drupal\sm\Transport\DrupalSql;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Database\Connection as CoreConnection;
use Symfony\Component\Messenger\Envelope;
use Symfony\Component\Messenger\Transport\Serialization\SerializerInterface;
use Symfony\Component\Messenger\Transport\TransportInterface;

/**
 * Transport for handling messages via Drupal's database connection.
 *
 * This class provides methods to send, acknowledge, and reject messages
 * using a Drupal SQL database as the transport mechanism. It utilizes
 * the `Sender` and `Receiver` classes to perform
 * the actual sending and receiving of messages.
 *
 * @phpstan-import-type DrupalSqlConfiguration from Connection
 * @internal
 */
final class Transport implements TransportInterface {

  private Connection $connection;
  private Receiver $receiver;
  private Sender $sender;

  /**
   * @phpstan-param DrupalSqlConfiguration $configuration
   */
  public function __construct(
    CoreConnection $connection,
    private readonly SerializerInterface $serializer,
    array $configuration,
    TimeInterface $time,
    MessageTableTracker $messageTableTracker,
  ) {
    $this->connection = new Connection($connection, $configuration, $time, $messageTableTracker);
  }

  public function get(): iterable {
    return $this->getReceiver()->get();
  }

  public function ack(Envelope $envelope): void {
    $this->getReceiver()->ack($envelope);
  }

  public function reject(Envelope $envelope): void {
    $this->getReceiver()->reject($envelope);
  }

  public function send(Envelope $envelope): Envelope {
    return $this->getSender()->send($envelope);
  }

  /**
   * Lazily create a receiver for this transport.
   */
  private function getReceiver(): Receiver {
    return $this->receiver ??= new Receiver(
      $this->connection,
      $this->serializer,
    );
  }

  /**
   * Lazily create a sender for this transport.
   */
  private function getSender(): Sender {
    return $this->sender ??= new Sender(
      $this->connection,
      $this->serializer,
    );
  }

}
