<?php

declare(strict_types=1);

namespace Drupal\Tests\sm\Kernel;

use Drupal\KernelTests\KernelTestBase;
use Drupal\sm_test\SmTestMessage;
use Drupal\Tests\sm\Traits\SmAsyncTrait;
use Drupal\Tests\sm\Traits\SmMappingTrait;
use Symfony\Component\Clock\MockClock;
use Symfony\Component\Messenger\Envelope;
use Symfony\Component\Messenger\Stamp\BusNameStamp;
use Symfony\Component\Messenger\Stamp\DelayStamp;
use Symfony\Component\Messenger\Worker;

/**
 * Tests the `asynchronous` (drupal-sql://) transport in sm.services.yml.
 */
final class SmDrupalSqlTransportTest extends KernelTestBase {

  use SmMappingTrait;
  use SmAsyncTrait;

  private string $time = '9am 10 Jan 2014 UTC';

  protected static $modules = [
    'sm',
    'sm_test',
  ];

  /**
   * Tests the message is handled asynchronously.
   */
  public function testDrupalSqlTransport(): void {
    $this->setRouting([
      SmTestMessage::class => 'sm.transport.asynchronous',
    ]);

    $asynchronousMessage = new SmTestMessage();
    static::bus()->dispatch($asynchronousMessage);

    // The message isn't handled yet.
    static::assertNull($asynchronousMessage->handledBy);

    // Check that the table exists and the message is stored.
    $messages = $this->queryForMessages();
    static::assertCount(1, $messages);

    // Check that the message is stored correctly.
    $envelope = new Envelope($asynchronousMessage);
    $envelope = $envelope->with(new BusNameStamp('sm.bus.default'));

    static::assertEquals(
      static::serializer()->encode($envelope)['body'],
      $messages[0]['body'],
    );
    static::assertEquals('default', $messages[0]['queue_name']);
    static::assertEquals([], \json_decode($messages[0]['headers']));
    static::assertEquals('2014-01-10 09:00:00', $messages[0]['created_at']);
    static::assertEquals('2014-01-10 09:00:00', $messages[0]['available_at']);
    static::assertNull($messages[0]['delivered_at']);

    $worker = new Worker(['asynchronous' => $this->receiver()],
      static::bus(),
      static::dispatcher(1),
      clock: new MockClock($this->time));
    $worker->run();

    // The message is handled.
    $messages = $this->queryForMessages();
    static::assertCount(0, $messages);
  }

  public function testDrupalSqlDelayedStamp(): void {
    $this->setRouting([
      SmTestMessage::class => 'sm.transport.asynchronous',
    ]);

    $envelope = Envelope::wrap(new SmTestMessage(), [new DelayStamp(1000)]);
    static::bus()->dispatch($envelope);

    $messages = $this->queryForMessages();
    // Current time, as simulated, plus one second.
    static::assertEquals('2014-01-10 09:00:01', $messages[0]['available_at']);
  }

  public function testDrupalSqlTableCleanup(): void {
    $this->setRouting([
      SmTestMessage::class => 'sm.transport.asynchronous',
    ]);

    $asynchronousMessage = new SmTestMessage();
    static::bus()->dispatch($asynchronousMessage);

    // Check that the table exists and the message is stored.
    $messages = $this->queryForMessages();
    static::assertCount(1, $messages);

    // Uninstall the module to drop the table.
    $this->container->get('module_installer')->uninstall(['sm']);
    static::assertFalse($this->tableExists('sm_messages'));
  }

}
