<?php

declare(strict_types=1);

namespace Drupal\Tests\sm\Traits;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Database\Connection as CoreConnection;
use Drupal\Core\DependencyInjection\ContainerBuilder;
use Drupal\sm\Transport\DrupalSql\Connection;
use Drupal\sm\Transport\DrupalSql\MessageTableTracker;
use Drupal\sm\Transport\DrupalSql\Transport;
use Drupal\sm_test\TimeMachine;
use Psr\EventDispatcher\EventDispatcherInterface;
use Symfony\Component\Messenger\EventListener\StopWorkerOnMessageLimitListener;
use Symfony\Component\Messenger\MessageBusInterface;
use Symfony\Component\Messenger\Transport\Receiver\ReceiverInterface;
use Symfony\Component\Messenger\Transport\Serialization\SerializerInterface;

/**
 * @phpstan-import-type DrupalSqlEnvelopeEncoded from \Drupal\sm\Transport\DrupalSql\Connection
 */
trait SmAsyncTrait {

  private string $currentTime = '9am 10 Jan 2014 UTC';

  /**
   * @phpstan-return list<DrupalSqlEnvelopeEncoded>
   */
  public function queryForMessages(): array {
    $select = static::connection()->select('sm_messages')
      ->fields('sm_messages')
      ->isNull('delivered_at')
      ->execute();

    return $select?->fetchAll(mode: \PDO::FETCH_ASSOC) ?? throw new \Exception('Failed to execute query.');
  }

  public function tableExists(string $name): bool {
    return static::connection()->schema()->tableExists($name);
  }

  public function receiver(): ReceiverInterface {
    $configuration = Connection::buildConfiguration('drupal-sql://default');

    return new Transport(
      static::connection(),
      static::serializer(),
      $configuration,
      static::time(),
      static::messageTableTracker(),
    );
  }

  public function register(ContainerBuilder $container): void {
    parent::register($container);

    $this->setSmRoutingToContainer($container);

    $container->getDefinition('datetime.time')
      ->setClass(TimeMachine::class)
      ->setArgument(0, $this->currentTime);
  }

  /**
   * A dispatcher that stops the worker after a certain number of messages.
   */
  protected function dispatcher(int $maximumNumberOfMessages): EventDispatcherInterface {
    /** @var \Symfony\Component\EventDispatcher\EventDispatcherInterface $eventDispatcher */
    $eventDispatcher = \Drupal::service('event_dispatcher');
    $eventDispatcher->addSubscriber(new StopWorkerOnMessageLimitListener($maximumNumberOfMessages));

    return $eventDispatcher;
  }

  protected static function connection(): CoreConnection {
    return \Drupal::database();
  }

  protected static function serializer(): SerializerInterface {
    return \Drupal::service('messenger.default_serializer');
  }

  protected static function time(): TimeInterface {
    return \Drupal::time();
  }

  protected static function messageTableTracker(): MessageTableTracker {
    return \Drupal::service(MessageTableTracker::class);
  }

  protected static function bus(): MessageBusInterface {
    /** @var \Symfony\Component\Messenger\MessageBusInterface */
    return \Drupal::service(MessageBusInterface::class);
  }

}
